#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
""" Pure-gateware, UTMI-compatible Full Speed PHY."""

from .phy import GatewarePHY
