# amaranth: UnusedElaboratable=no
#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
""" USB3 Physical-Layer modules """

from .layer import USB3PhysicalLayer

__all__ = ['USB3PhysicalLayer']
